SystemOrganization addCategory: #AndreasProfiler!

Object subclass: #AndreasSystemProfiler
	instanceVariableNames: 'semaphore ticks profilerProcess tallyRoot vmStats totalTally totalTime startTime ticksPerMSec totalTicks'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AndreasProfiler'!

!AndreasSystemProfiler commentStamp: 'eem 8/2/2013 13:45' prior: 0!
AndreasSystemProfiler instances observe and report the amount of time spent in methods.  AndreasSystemProfiler uses sub-msec VM supported PC sampling.  AndreasSystemProfiler is a plug-in replacement for MessageTally.  It is more accurate because it uses VM support to tell it which primtiive was running when it took a sample.  MessageTally simply ascribes a primitive's cost to the method at the next suspendsion point, which, in some contexts, can yield wildly misleading results.

	AndreasSystemProfiler spyOn:  [20 timesRepeat: 
			[Transcript show: 100 factorial printString]]

NOTE: a higher-level user interface (combining the MessageTally result tree with a method browser) is available from TimeProfileBrowser.  But you'll have to plumb AndreasSystemProfiler into TimeProfileBrowser yourself.

	TimeProfileBrowser spyOn:  [20 timesRepeat: 
			[Transcript show: 100 factorial printString]]

	In Memory of Andreas Raab.  Author, Friend, Colleague. 	http://forum.world.st/In-Memory-of-Andreas-Raab-td4663424.html
	Released by Ron, Julie and David!

!AndreasSystemProfiler class methodsFor: 'LICENSE' stamp: 'RJT 1/23/2013 15:34'!
LICENSE
	^'Project Squeak

	In Memory of Andreas Raab.  Author, Friend, Colleague. 	http://forum.world.st/In-Memory-of-Andreas-Raab-td4663424.html
	Copyright (c) 2005-2013, 3D Immersive Collaboration Consulting, LLC., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

Licensed under MIT License (MIT)
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.'! !

!AndreasSystemProfiler class methodsFor: 'testing' stamp: 'jmv 4/16/2013 23:32'!
canWork
	"
	AndreasSystemProfiler canWork
	"
	Smalltalk highResClock = 0 ifTrue: [ ^false ].
	[
		Smalltalk profileSemaphore: nil.
		Smalltalk profileStart: 0 ] on: Error do: [ ^false ].
	^true! !

!AndreasSystemProfiler class methodsFor: 'spying' stamp: 'RJT 1/23/2013 15:28'!
default
	^self new! !

!AndreasSystemProfiler class methodsFor: 'spying' stamp: 'eem 8/2/2013 13:53'!
spyFor: seconds
	"Run the system profiler for the specified number of seconds"
	"AndreasSystemProfiler spyFor: 5"
	^self spyOn:[(Delay forSeconds: seconds) wait].! !

!AndreasSystemProfiler class methodsFor: 'spying' stamp: 'eem 8/2/2013 13:53'!
spyOn: aBlock
	"AndreasSystemProfiler spyOn:[10000 timesRepeat: [3.14159 printString]]"
	| profiler |
	profiler := self new.
	^[profiler spyOn: aBlock] ensure: [profiler doReport]! !

!AndreasSystemProfiler class methodsFor: 'spying' stamp: 'RJT 1/23/2013 15:28'!
spyOnWorldFor: seconds
	"Run the system profiler for the specified number of seconds,
	 spying on the morphic world in which it was launched. 
	 Handy for running Workspace do-its."

	^self spyOn: [  | deadline |
		deadline := Time totalSeconds + seconds.
		[Time totalSeconds < deadline]
			whileTrue: [ World doOneCycle ]].! !

!AndreasSystemProfiler methodsFor: 'reporting' stamp: 'RJT 1/23/2013 15:28'!
doReport
	"Report the results of this profiler run"
	UIManager default edit: self report label: 'Spy Results'.! !

!AndreasSystemProfiler methodsFor: 'reporting' stamp: 'RJT 1/23/2013 15:28'!
report
	"Answer a report, with cutoff percentage of each element of the tree"
	^String streamContents:[:s| self report: s].! !

!AndreasSystemProfiler methodsFor: 'reporting' stamp: 'RJT 1/23/2013 15:28'!
report: strm 
	"Print a report, with cutoff percentage of each element of the tree 
	(leaves, roots, tree)=2, on the stream, strm."

	self report: strm cutoff: 1! !

!AndreasSystemProfiler methodsFor: 'reporting' stamp: 'RJT 1/23/2013 15:28'!
report: strm cutoff: threshold 
	tallyRoot tally = 0 ifTrue:[
		strm nextPutAll: ' - no tallies obtained'
	] ifFalse:[
		strm nextPutAll: 'Reporting - ', totalTally asStringWithCommas,' tallies, ',
							 totalTime asStringWithCommas, ' msec.'; cr; cr.
		tallyRoot fullPrintOn: strm tallyExact: false orThreshold: threshold time: totalTime.
	].
	totalTime isZero ifFalse:[
		self reportGCStatsOn: strm.
		self reportProcessStatsOn: strm.
	].! !

!AndreasSystemProfiler methodsFor: 'reporting' stamp: 'ul 10/21/2014 17:01'!
reportGCStatsOn: str
	| oldSpaceEnd youngSpaceEnd memoryEnd fullGCs fullGCTime incrGCs incrGCTime tenureCount upTime rootOverflows |
	upTime := totalTime.
	oldSpaceEnd			:= vmStats at: 1.
	youngSpaceEnd		:= vmStats at: 2.
	memoryEnd			:= vmStats at: 3.
	fullGCs				:= vmStats at: 7.
	fullGCTime			:= vmStats at: 8.
	incrGCs				:= vmStats at: 9.
	incrGCTime			:= vmStats at: 10.
	tenureCount			:= vmStats at: 11.
	rootOverflows		:= vmStats at: 22.

	str cr.
	str	nextPutAll: '**Memory**'; cr.
	str	nextPutAll:	'	old			';
		nextPutAll: (oldSpaceEnd asStringWithCommasSigned: true); nextPutAll: ' bytes'; cr.
	str	nextPutAll: '	young		';
		nextPutAll: ((youngSpaceEnd - oldSpaceEnd) asStringWithCommasSigned: true); nextPutAll: ' bytes'; cr.
	str	nextPutAll: '	used		';
		nextPutAll: (youngSpaceEnd asStringWithCommasSigned: true); nextPutAll: ' bytes'; cr.
	str	nextPutAll: '	free		';
		nextPutAll: ((memoryEnd - youngSpaceEnd) asStringWithCommasSigned: true); nextPutAll: ' bytes'; cr.

	str cr.
	str	nextPutAll: '**GCs**'; cr.
	str	nextPutAll: '	full			';
		print: fullGCs; nextPutAll: ' totalling '; nextPutAll: fullGCTime asStringWithCommas; nextPutAll: 'ms (';
		print: ((fullGCTime / upTime * 100) roundTo: 1.0);
		nextPutAll: '% uptime)'.
	fullGCs = 0 ifFalse:
		[str	nextPutAll: ', avg '; print: ((fullGCTime / fullGCs) roundTo: 1.0); nextPutAll: 'ms'].
	str	cr.
	str	nextPutAll: '	incr		';
		print: incrGCs; nextPutAll: ' totalling '; nextPutAll: incrGCTime asStringWithCommas; nextPutAll: 'ms (';
		print: ((incrGCTime / upTime * 100) roundTo: 1.0);
		nextPutAll: '% uptime)'.
	incrGCs = 0 ifFalse:
		[str nextPutAll:', avg '; print: ((incrGCTime / incrGCs) roundTo: 1.0); nextPutAll: 'ms'].
	str cr.
	str	nextPutAll: '	tenures		';
		nextPutAll: tenureCount asStringWithCommas.
	tenureCount = 0 ifFalse:
		[str nextPutAll: ' (avg '; print: (incrGCs / tenureCount) asInteger; nextPutAll: ' GCs/tenure)'].
	str	cr.
	str	nextPutAll: '	root table	';
		nextPutAll: rootOverflows asStringWithCommas; nextPutAll:' overflows'.
	str cr.
! !

!AndreasSystemProfiler methodsFor: 'reporting' stamp: 'eem 11/26/2019 12:25'!
reportProcessStatsOn: str
	| totalSwitches pageOverflows pageDivorces actualSwitches |
	vmStats size >= 61 ifFalse:[^self]. "don't try this on the closure VM"
	totalSwitches := vmStats at: 56.
	actualSwitches := totalSwitches - (2*totalTally). "exclude to/from profiler"
	pageOverflows := vmStats at: 60.
	pageDivorces := vmStats at: 61.

	str cr; nextPutAll: '**Processes**'; cr.
	str tab; nextPutAll: 'Total process switches: '; print: totalSwitches; cr.
	str tab; nextPutAll: 'Without Profiler: '; print: actualSwitches; cr.
	str tab; nextPutAll: 'Stack page overflows: '; print: pageOverflows; cr.
	str tab; nextPutAll: 'Stack page divorces: '; print: pageDivorces; cr! !

!AndreasSystemProfiler methodsFor: 'profiling' stamp: 'eem 8/2/2013 13:49'!
runProfilerProcess
	"Run the profiler process"
	| process tallyStart tallyTicks tallyStop method leaf |
	tallyRoot := ASPTally new class: nil method: nil.
	totalTally := 0.
	Smalltalk profileSemaphore: semaphore.
	totalTicks := 0.
	tallyStart := tallyStop := Smalltalk highResClock.
	[true] whileTrue:[
		tallyStart := tallyStop.
		tallyStart := Smalltalk highResClock.
		Smalltalk profileStart: ticks. "run for n ticks"
		semaphore wait.
		tallyStop := Smalltalk highResClock.
		tallyTicks := tallyStop - tallyStart.
		totalTicks := totalTicks + tallyTicks.
		process := Smalltalk profileSample.
		method := Smalltalk profilePrimitive.
		totalTally := totalTally + 1.
		process ifNotNil:[
			leaf := tallyRoot tally: (process suspendedContext ifNil:[thisContext]) by: tallyTicks.
			method ifNotNil:[leaf tallyMethod: method by: tallyTicks].
		].
	].
! !

!AndreasSystemProfiler methodsFor: 'profiling' stamp: 'fbs 2/8/2013 22:43'!
spyOn: aBlock
	"Profile system activity during execution of aBlock. 
	The argument is the desired samples per *milli* second.
	Mostly for polymorphism with MessageTally."
	self startProfiling.
	^ aBlock ensure:[self stopProfiling].
! !

!AndreasSystemProfiler methodsFor: 'profiling' stamp: 'eem 10/5/2020 22:04'!
startProfiling
	"Start the profiler process taking samplesPerMsec samples per *milli* second"
	| t0 |
	semaphore := Semaphore new.
	"Run a 10 msec busy loop to figure out the ticks per msec"
	"First step to the next clock granule"
	t0 := Time utcMicrosecondClock + 1000.
	[Time utcMicrosecondClock >= t0] whileFalse.
	"Now run for 10 ms"
	ticksPerMSec := Smalltalk highResClock.
	[Time utcMicrosecondClock >= (t0 + 10000)] whileFalse.
	ticksPerMSec := (Smalltalk highResClock - ticksPerMSec) 
						// (Time utcMicrosecondClock - t0).
	"Try to get 10 samples per msec"
	ticks := ticksPerMSec // 10.
	startTime := Time utcMicrosecondClock.
	vmStats := Smalltalk getVMParameters.
	profilerProcess := [self runProfilerProcess] forkAt: Processor timingPriority - 1

	"How granular is the microsecond clock?  Not very..."
	"| times st t nt |
	times := OrderedCollection with: (st := t := Time utcMicrosecondClock).
	[times size < 100] whileTrue:
		[[(nt := Time utcMicrosecondClock) = t] whileTrue.
		 times addLast: (t := nt)].
	times collect: [:time| time - st]"! !

!AndreasSystemProfiler methodsFor: 'profiling' stamp: 'eem 10/5/2020 21:58'!
stopProfiling
	"Stop the profiler process"
	Smalltalk profileSemaphore: nil.
	Smalltalk profileStart: 0. "<- profile stops now"
	totalTime := (Time utcMicrosecondClock - startTime) + 500 // 1000.
	Smalltalk getVMParameters keysAndValuesDo:
		[:idx :value| 
		value isInteger ifTrue: [vmStats at: idx put: (value - ((vmStats at: idx) ifNil: [0]))]].
	profilerProcess ifNotNil:
		[profilerProcess suspend.
		 profilerProcess := nil]
! !

!SmalltalkImage methodsFor: '*AndreasProfiler-profiling' stamp: 'ar 6/22/2007 15:25'!
highResClock	"Smalltalk highResClock"
	"Primitive. Answer the value of the high resolution clock if this computer has one.
	Usually, this should be the highest resolution value available, for example on Intel
	it will be the value of the time stamp counter register."
	<primitive: 'primitiveHighResClock'>
	^0! !

!SmalltalkImage methodsFor: '*AndreasProfiler-profiling' stamp: 'ar 6/22/2007 15:24'!
profilePrimitive
	"Primitive. Answer the primitive sample by the profiler, or nil if the profiler isn't active.
	See also primitiveProfileStart."
	<primitive: 'primitiveProfilePrimitive'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: '*AndreasProfiler-profiling' stamp: 'ar 6/22/2007 15:24'!
profileSample
	"Primitive. Answer the last sample taken by the profiler, or nil if the profiler isn't active.
	See also primitiveProfileStart."
	<primitive: 'primitiveProfileSample'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: '*AndreasProfiler-profiling' stamp: 'ar 6/22/2007 15:24'!
profileSemaphore: aSemaphore
	"Primitive. Install the semaphore to be used for profiling, 
	or nil if no semaphore should be used.
	See also primitiveProfileStart."
	<primitive: 'primitiveProfileSemaphore'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: '*AndreasProfiler-profiling' stamp: 'ar 6/22/2007 15:24'!
profileStart: counter
	"Primitive. Begin profiling execution every by using the interrupt check-counter instead of a time-based process (which is limited to timing resolution and triggers off the same signal that many of the processes being profiled trigger off leading to consistently wrong results).
	The argument is the number of interrupt checks (method activations) to let go by before taking a sample. The sample is being stored in the profileSample iVar which can be retrieved by executing primitiveProfileSample. When a sample is taken, it signals the semaphore specified in primitiveProfileSemaphore.
	If the argument is less or equal to zero, it disables profiling."
	<primitive: 'primitiveProfileStart'>
	^self primitiveFailed! !

Link subclass: #ASPTally
	instanceVariableNames: 'class method receivers tally senders blockNesting'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AndreasProfiler'!

!ASPTally commentStamp: '<historical>' prior: 0!
Tally for assembling system profiles. It's a subclass of Link so we can chain tallies together in the receivers list.!

!ASPTally class methodsFor: 'LICENSE' stamp: 'RJT 1/23/2013 15:34'!
LICENSE
	^'Project Squeak

	In Memory of Andreas Raab.  Author, Friend, Colleague. 	http://forum.world.st/In-Memory-of-Andreas-Raab-td4663424.html
	Copyright (c) 2005-2013, 3D Immersive Collaboration Consulting, LLC., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

Licensed under MIT License (MIT)
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.'! !

!ASPTally methodsFor: 'converting' stamp: 'eem 6/27/2015 12:10'!
asArray
	^Array streamContents:
		[:s| | link |
		link := self.
		[link == nil] whileFalse:
			[s nextPut: link.
			 link := link nextLink]]! !

!ASPTally methodsFor: 'converting' stamp: 'ar 6/11/2007 22:34'!
asSortedCollection: aCollection
	"Create a sorted collection from the given input"
	^aCollection asSortedCollection:[:tA :tB| tA tally >= tB tally].! !

!ASPTally methodsFor: 'accessing' stamp: 'eem 1/27/2015 14:17'!
blockNestingCount
	^blockNesting! !

!ASPTally methodsFor: 'tallying' stamp: 'eem 1/27/2015 14:12'!
blockNestingCountOf: context
	| count nest |
	count := 0.
	nest := context.
	[nest closure notNil] whileTrue:
		[count := count + 1.
		 nest := nest closure outerContext].
	^count! !

!ASPTally methodsFor: 'report' stamp: 'ar 6/11/2007 22:31'!
bump: hitCount fromSender: senderTally
	"Add this hitCount to the total, and include a reference to the
	sender responsible for the increment"
	self bumpBy: hitCount.
	senders ifNil:[senders := OrderedCollection new].
	senderTally ifNotNil:[senders add: (senderTally copyWithTally: hitCount)]! !

!ASPTally methodsFor: 'tallying' stamp: 'ar 6/11/2007 21:54'!
bumpBy: count
	"Bump this tally by the specified amount"
	tally := tally + count! !

!ASPTally methodsFor: 'initialize' stamp: 'eem 1/27/2015 14:08'!
class: aClass method: aCompiledMethod
	class := aClass.
	method := aCompiledMethod.
	tally := blockNesting := 0! !

!ASPTally methodsFor: 'initialize' stamp: 'eem 1/27/2015 14:08'!
class: aClass method: aCompiledMethod nesting: blockNestingCount
	class := aClass.
	method := aCompiledMethod.
	blockNesting := blockNestingCount.
	tally := 0! !

!ASPTally methodsFor: 'copying' stamp: 'eem 8/2/2013 13:47'!
copyWithTally: hitCount
	^(self class new class: class method: method) bumpBy: hitCount! !

!ASPTally methodsFor: 'report' stamp: 'ar 6/11/2007 22:28'!
fullPrintOn: aStream tallyExact: isExact orThreshold: perCent time: totalTime
	| threshold |  
	isExact ifFalse: [threshold := (perCent asFloat / 100 * tally) rounded].
	aStream nextPutAll: '**Tree**'; cr.
	self treePrintOn: aStream
		tabs: OrderedCollection new
		thisTab: ''
		total: tally
		totalTime: totalTime
		tallyExact: isExact
		orThreshold: threshold.
	aStream nextPut: Character newPage; cr.
	aStream nextPutAll: '**Leaves**'; cr.
	self leavesPrintOn: aStream
		tallyExact: isExact
		orThreshold: threshold
		time: totalTime.! !

!ASPTally methodsFor: 'report' stamp: 'eem 8/2/2013 13:47'!
into: leafDict fromSender: senderTally
	| leafNode |
	leafNode := leafDict
					at: method
					ifAbsentPut: [self class new class: class method: method].
	leafNode bump: tally fromSender: senderTally! !

!ASPTally methodsFor: 'testing' stamp: 'ar 6/11/2007 22:50'!
isPrimitives
	"Detect pseudo node used to carry tally of local hits"
	^ method == nil! !

!ASPTally methodsFor: 'report' stamp: 'ar 6/11/2007 22:29'!
leavesInto: leafDict fromSender: senderTally
	| rcvrs |
	rcvrs := self sonsOver: 0.
	rcvrs size = 0
		ifTrue: [self into: leafDict fromSender: senderTally]
		ifFalse: [rcvrs do:[:node |
				node isPrimitives
					ifTrue: [node leavesInto: leafDict fromSender: senderTally]
					ifFalse: [node leavesInto: leafDict fromSender: self]]]! !

!ASPTally methodsFor: 'report' stamp: 'ar 6/11/2007 22:35'!
leavesPrintOn: aStream tallyExact: isExact orThreshold: threshold time: totalTime
	| dict |
	dict := IdentityDictionary new: 100.
	self leavesInto: dict fromSender: nil.
	isExact ifTrue:[
		(self asSortedCollection: dict) do: [:node |
			node printOn: aStream total: tally totalTime: nil tallyExact: isExact.
			node printSenderCountsOn: aStream].
	] ifFalse:[
		((self asSortedCollection: dict)
			select: [:node | node tally > threshold]) do: [:node |
				node printOn: aStream total: tally totalTime: totalTime tallyExact: isExact].
	].! !

!ASPTally methodsFor: 'accessing' stamp: 'ar 6/11/2007 22:24'!
maxClassNameSize
	"Return the default maximum width of the class name alone"
	^30! !

!ASPTally methodsFor: 'accessing' stamp: 'ar 6/11/2007 22:24'!
maxClassPlusSelectorSize
	"Return the default maximum width of the class plus selector together (not counting the '>>')"
	^60! !

!ASPTally methodsFor: 'accessing' stamp: 'bgf 3/6/2008 12:21'!
maxTabs
	"Return the default number of tabs after which leading white space is compressed"
	^36! !

!ASPTally methodsFor: 'accessing' stamp: 'ar 6/11/2007 21:57'!
method
	"Answer the CompiledMethod associated with this tally"
	^method! !

!ASPTally methodsFor: 'initialize' stamp: 'ar 6/11/2007 22:50'!
primitives: anInteger
	"Make the receiver be a node of unassigned primitives"
	tally := anInteger.
	method := nil. "indicates primitives"! !

!ASPTally methodsFor: 'report' stamp: 'eem 1/27/2015 14:14'!
printOn: aStream
	blockNesting ifNotNil:
		[blockNesting timesRepeat:
			[aStream nextPutAll: '[] in ']].
	aStream print: class; nextPutAll: '>>'; print: (method ifNotNil:[method selector storeString]).
	aStream nextPutAll: ' -- '; print: tally.! !

!ASPTally methodsFor: 'report' stamp: 'eem 1/27/2015 14:27'!
printOn: aStream total: total totalTime: totalTime tallyExact: isExact 
	| className myTally aClass percentage |
	isExact
		ifTrue:
			[myTally := tally.
			receivers ifNotNil:[receivers asArray do: [:r | myTally := myTally - r tally]].
			aStream print: myTally; space]
		ifFalse:
			[(percentage := tally asFloat / total * 100.0) printOn: aStream maxDecimalPlaces: 2.
			aStream
				nextPutAll: ' (';
				nextPutAll: (percentage * totalTime / 100) rounded asStringWithCommas;
				nextPutAll: ')  '].
	self isPrimitives
		ifTrue:
			[aStream nextPutAll: 'primitives'; cr]
		ifFalse:
			[aClass := method methodClass.
			className := aClass name contractTo: self maxClassNameSize.
			blockNesting > 0 ifTrue:
				[aStream
					next: blockNesting put: $[;
					next: blockNesting put: $];
					space].
			aStream
				nextPutAll: class name;
				nextPutAll: (aClass = class 
							ifTrue: [' ']
							ifFalse: ['(' , aClass name , ') ']);
				nextPutAll: (method selector
								contractTo: self maxClassPlusSelectorSize - className size);
				cr]! !

!ASPTally methodsFor: 'report' stamp: 'ar 2/27/2008 14:09'!
sonsOver: threshold
	"Answer the sons with tallys over the given threshold"
	receivers ifNil:[^#()].
	^receivers asArray select:[:son | son tally > threshold].! !

!ASPTally methodsFor: 'accessing' stamp: 'ar 6/11/2007 21:58'!
tally
	"Answer the tally count for this node"
	^tally! !

!ASPTally methodsFor: 'tallying' stamp: 'ul 1/27/2015 22:16'!
tally: context by: count
	"Explicitly tally the specified context and its stack."
	
	^(context sender 
		ifNil: [ self bumpBy: count ]
		ifNotNil: [ :root | self tally: root by: count ]) tallyPath: context by: count! !

!ASPTally methodsFor: 'tallying' stamp: 'eem 8/2/2013 13:48'!
tallyMethod: aMethod by: count
	"Called explicitly and needs to decrement receiver's tally count"
	| node |
	node := receivers.
	[node == nil] whileFalse:
		[node method == aMethod ifTrue:[^node bumpBy: count].
		node := node nextLink].
	node := self class new class: aMethod methodClass method: aMethod.
	node nextLink: receivers.
	receivers := node.
	^node bumpBy: count! !

!ASPTally methodsFor: 'tallying' stamp: 'eem 1/27/2015 14:18'!
tallyPath: context by: count
	"Tally the context chain"
	| aMethod nesting theTally |
	aMethod := context method.
	nesting := self blockNestingCountOf: context.
	theTally := receivers.
	[theTally == nil] whileFalse:
		[(theTally method == aMethod
		  and: [theTally blockNestingCount = nesting]) ifTrue:[^theTally bumpBy: count].
		theTally := theTally nextLink].
	theTally := self class new
				class: context receiver class
				method: aMethod
				nesting: nesting.
	theTally nextLink: receivers.
	receivers := theTally.
	^theTally bumpBy: count! !

!ASPTally methodsFor: 'report' stamp: 'ar 6/11/2007 22:50'!
treePrintOn: aStream tabs: tabs thisTab: myTab total: total totalTime: totalTime tallyExact: isExact orThreshold: threshold 
	| sons sonTab |
	tabs do: [:tab | aStream nextPutAll: tab].
	tabs size > 0 
		ifTrue:[self printOn: aStream total: total totalTime: totalTime tallyExact: isExact].
	sons := isExact ifTrue: [receivers] ifFalse: [self sonsOver: threshold].
	sons isEmpty ifFalse:[
		tabs addLast: myTab.
		sons := self asSortedCollection: sons.
		1 to: sons size do:[:i | 
			sonTab := i < sons size ifTrue: ['  |'] ifFalse: ['  '].
			(sons at: i) treePrintOn: aStream
						tabs: (tabs size < self maxTabs 
								ifTrue: [tabs]
								ifFalse: [(tabs select: [:x | x = '[']) copyWith: '['])
						thisTab: sonTab
						total: total
						totalTime: totalTime
						tallyExact: isExact
						orThreshold: threshold].
			tabs removeLast]! !
